from typing import Hashable, Generator, Optional, Iterable
import time

import pandas as pd
import sqlalchemy
from sqlalchemy.pool import NullPool
from sqlalchemy.sql import select
from sqlalchemy import and_, or_, func

from .backend import Backend

_DEFAULT_SQL_URL = "sqlite:///"
_DEFAULT_SQL_STR_LEN = 64


class SQLBackend(Backend):
    """
    A graph datastore that uses a SQL-like store for persistance and queries.

    """

    def __init__(
        self,
        directed: bool = False,
        node_table_name: str = None,
        edge_table_name: str = None,
        db_url: str = _DEFAULT_SQL_URL,
        primary_key: str = "ID",
        sqlalchemy_kwargs: dict = None,
    ) -> None:
        """
        Create a new SQL-backed graph store.

        Arguments:
            node_table_name (str: "grand_Nodes"): The name to use for the node
                table in DynamoDB.
            edge_table_name (str: "grand_Edges"): The name to use for the edge
                table in DynamoDB.
            db_url (str: _DEFAULT_SQL_URL): The URL to use for the SQL db.
            primary_key (str: "ID"): The default primary key to use for the
                tables. Note that this key cannot exist in your metadata dicts.

        """
        self._directed = directed
        self._node_table_name = node_table_name or "grand_Nodes"
        self._edge_table_name = edge_table_name or "grand_Edges"

        self._primary_key = primary_key
        self._edge_source_key = "Source"
        self._edge_target_key = "Target"

        sqlalchemy_kwargs = sqlalchemy_kwargs or {}
        self._engine = sqlalchemy.create_engine(db_url, **sqlalchemy_kwargs)
        self._connection = self._engine.connect()
        self._metadata = sqlalchemy.MetaData()

        if not self._engine.dialect.has_table(self._connection, self._node_table_name):
            self._node_table = sqlalchemy.Table(
                self._node_table_name,
                self._metadata,
                sqlalchemy.Column(
                    self._primary_key,
                    sqlalchemy.String(_DEFAULT_SQL_STR_LEN),
                    primary_key=True,
                ),
                sqlalchemy.Column("_metadata", sqlalchemy.JSON),
            )
            self._node_table.create(self._engine)
        else:
            self._node_table = sqlalchemy.Table(
                self._node_table_name,
                self._metadata,
                autoload=True,
                autoload_with=self._engine,
            )

        if not self._engine.dialect.has_table(self._connection, self._edge_table_name):
            self._edge_table = sqlalchemy.Table(
                self._edge_table_name,
                self._metadata,
                sqlalchemy.Column(
                    self._primary_key,
                    sqlalchemy.String(_DEFAULT_SQL_STR_LEN),
                    primary_key=True,
                ),
                sqlalchemy.Column("_metadata", sqlalchemy.JSON),
                sqlalchemy.Column(
                    self._edge_source_key, sqlalchemy.String(_DEFAULT_SQL_STR_LEN)
                ),
                sqlalchemy.Column(
                    self._edge_target_key, sqlalchemy.String(_DEFAULT_SQL_STR_LEN)
                ),
            )
            self._edge_table.create(self._engine)
        else:
            self._edge_table = sqlalchemy.Table(
                self._edge_table_name,
                self._metadata,
                autoload=True,
                autoload_with=self._engine,
            )

    # def __del__(self):
    # self._connection.close()

    def is_directed(self) -> bool:
        """
        Return True if the backend graph is directed.

        Arguments:
            None

        Returns:
            bool: True if the backend graph is directed.

        """
        return self._directed

    def teardown(self, yes_i_am_sure: bool = False):
        """
        Tear down this graph, deleting all evidence it once was here.

        """
        if yes_i_am_sure:
            self._node_table.drop(self._engine)
            self._edge_table.drop(self._engine)

    def add_node(self, node_name: Hashable, metadata: dict) -> Hashable:
        """
        Add a new node to the graph.

        Insert a new document into the nodes table.

        Arguments:
            node_name (Hashable): The ID of the node
            metadata (dict: None): An optional dictionary of metadata

        Returns:
            Hashable: The ID of this node, as inserted

        """
        self._connection.execute(
            self._node_table.insert(),
            **{self._primary_key: node_name, "_metadata": metadata},
        )
        return node_name

    def all_nodes_as_iterable(self, include_metadata: bool = False) -> Generator:
        """
        Get a generator of all of the nodes in this graph.

        Arguments:
            include_metadata (bool: False): Whether to include node metadata in
                the response

        Returns:
            Generator: A generator of all nodes (arbitrary sort)

        """
        results = self._connection.execute(self._node_table.select()).fetchall()
        if include_metadata:
            return [(row[self._primary_key], row["_metadata"]) for row in results]
        return [row[self._primary_key] for row in results]

    def has_node(self, u: Hashable) -> bool:
        """
        Return true if the node exists in the graph.

        Arguments:
            u (Hashable): The ID of the node to check

        Returns:
            bool: True if the node exists
        """
        return len(
            self._connection.execute(
                self._node_table.select().where(
                    self._node_table.c[self._primary_key] == u
                )
            ).fetchall()
        )

    def add_edge(self, u: Hashable, v: Hashable, metadata: dict):
        """
        Add a new edge to the graph between two nodes.

        If the graph is directed, this edge will start (source) at the `u` node
        and end (target) at the `v` node.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID
            metadata (dict): Optional metadata to associate with the edge

        Returns:
            Hashable: The edge ID, as inserted.

        """
        pk = f"__{u}__{v}"

        if not self.has_node(u):
            self.add_node(u, {})
        if not self.has_node(v):
            self.add_node(v, {})

        try:
            self._connection.execute(
                self._edge_table.insert(),
                **{
                    self._primary_key: pk,
                    self._edge_source_key: u,
                    self._edge_target_key: v,
                    "_metadata": metadata,
                },
            )
        except sqlalchemy.exc.IntegrityError:
            # Edge already exists
            pass
        return pk

    def all_edges_as_iterable(self, include_metadata: bool = False) -> Generator:
        """
        Get a list of all edges in this graph, arbitrary sort.

        Arguments:
            include_metadata (bool: False): Whether to include edge metadata

        Returns:
            Generator: A generator of all edges (arbitrary sort)

        """
        return iter(
            [
                (e.Source, e.Target, e._metadata)
                if include_metadata
                else (e.Source, e.Target)
                for e in self._connection.execute(self._edge_table.select()).fetchall()
            ]
        )

    def get_node_by_id(self, node_name: Hashable):
        """
        Return the data associated with a node.

        Arguments:
            node_name (Hashable): The node ID to look up

        Returns:
            dict: The metadata associated with this node

        """
        res = (
            self._connection.execute(
                self._node_table.select().where(
                    self._node_table.c[self._primary_key] == node_name
                )
            )
            .fetchone()
            ._metadata
        )
        return res

    def get_edge_by_id(self, u: Hashable, v: Hashable):
        """
        Get an edge by its source and target IDs.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID

        Returns:
            dict: Metadata associated with this edge

        """
        if self._directed:
            pk = f"__{u}__{v}"
            return (
                self._connection.execute(
                    self._edge_table.select().where(
                        self._edge_table.c[self._primary_key] == pk
                    )
                )
                .fetchone()
                ._metadata
            )
        else:
            return (
                self._connection.execute(
                    self._edge_table.select().where(
                        or_(
                            (self._edge_table.c[self._primary_key] == f"__{u}__{v}"),
                            (self._edge_table.c[self._primary_key] == f"__{v}__{u}"),
                        )
                    )
                )
                .fetchone()
                ._metadata
            )

    def get_node_neighbors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        """
        Get a generator of all downstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        if self._directed:
            res = self._connection.execute(
                self._edge_table.select().where(
                    self._edge_table.c[self._edge_source_key] == u
                )
            ).fetchall()
        else:
            res = self._connection.execute(
                self._edge_table.select().where(
                    or_(
                        (self._edge_table.c[self._edge_source_key] == u),
                        (self._edge_table.c[self._edge_target_key] == u),
                    )
                )
            ).fetchall()

        if include_metadata:
            return {
                (
                    r[self._edge_source_key]
                    if r[self._edge_source_key] != u
                    else r[self._edge_target_key]
                ): r["_metadata"]
                for r in res
            }

        return iter(
            [
                (
                    r[self._edge_source_key]
                    if r[self._edge_source_key] != u
                    else r[self._edge_target_key]
                )
                for r in res
            ]
        )

    def get_node_predecessors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        """
        Get a generator of all upstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        if self._directed:
            res = self._connection.execute(
                self._edge_table.select().where(
                    self._edge_table.c[self._edge_target_key] == u
                )
            ).fetchall()
        else:
            res = self._connection.execute(
                self._edge_table.select().where(
                    or_(
                        (self._edge_table.c[self._edge_target_key] == u),
                        (self._edge_table.c[self._edge_source_key] == u),
                    )
                )
            ).fetchall()

        if include_metadata:
            return {
                (
                    r[self._edge_source_key]
                    if r[self._edge_source_key] != u
                    else r[self._edge_target_key]
                ): r["_metadata"]
                for r in res
            }

        return iter(
            [
                (
                    r[self._edge_source_key]
                    if r[self._edge_source_key] != u
                    else r[self._edge_target_key]
                )
                for r in res
            ]
        )

    def get_node_count(self) -> Iterable:
        """
        Get an integer count of the number of nodes in this graph.

        Arguments:
            None

        Returns:
            int: The count of nodes

        """
        return self._connection.execute(
            select([func.count()]).select_from(self._node_table)
        ).scalar()

    def ingest_from_edgelist_dataframe(
        self, edgelist: pd.DataFrame, source_column: str, target_column: str
    ) -> None:
        """
        Ingest an edgelist from a Pandas DataFrame.

        """
        # Produce edge list:

        edge_tic = time.time()
        newlist = edgelist.rename(
            columns={
                source_column: self._edge_source_key,
                target_column: self._edge_target_key,
            }
        )

        newlist[self._primary_key] = edgelist.apply(
            lambda x: f"__{x[source_column]}__{x[target_column]}", axis="columns"
        )
        newlist["_metadata"] = edgelist.apply(
            lambda x: {
                k: v for k, v in x.items() if k not in [source_column, target_column]
            },
            axis="columns",
        )

        newlist[
            [
                self._edge_source_key,
                self._edge_target_key,
                self._primary_key,
                "_metadata",
            ]
        ].to_sql(
            self._edge_table_name,
            self._engine,
            index=False,
            if_exists="append",
            dtype={"_metadata": sqlalchemy.JSON},
        )

        edge_toc = time.time() - edge_tic

        # now ingest nodes:
        node_tic = time.time()
        nodes = edgelist[source_column].append(edgelist[target_column]).unique()
        pd.DataFrame(
            [
                {
                    self._primary_key: node,
                    # no metadata:
                    "_metadata": {},
                }
                for node in nodes
            ]
        ).to_sql(
            self._node_table_name,
            self._engine,
            index=False,
            if_exists="replace",
            dtype={"_metadata": sqlalchemy.JSON},
        )

        return {
            "node_count": len(nodes),
            "node_duration": time.time() - node_tic,
            "edge_count": len(edgelist),
            "edge_duration": edge_toc,
        }




from typing import Hashable, Generator, Optional, Iterable
import time
import concurrent.futures

import pandas as pd
import boto3
from boto3.dynamodb.conditions import Key

from .backend import Backend


_DEFAULT_DYNAMODB_URL = "http://localhost:4566"
_N_PARALLEL_REQUESTS = 16


def _dynamo_table_exists(table_name: str, client: boto3.client):
    """
    Check to see if the DynamoDB table already exists.

    Returns:
        bool: Whether table exists

    """
    existing_tables = client.list_tables()["TableNames"]
    return table_name in existing_tables


def _create_dynamo_table(
    table_name: str, primary_key: str, client, read_write_units: Optional[int] = None,
):
    if read_write_units is not None:
        raise NotImplementedError("Non-on-demand billing is not currently supported.")

    return client.create_table(
        TableName=table_name,
        KeySchema=[
            {"AttributeName": primary_key, "KeyType": "HASH"},  # Partition key
            # {"AttributeName": "title", "KeyType": "RANGE"},  # Sort key
        ],
        AttributeDefinitions=[
            {"AttributeName": primary_key, "AttributeType": "S"},
            # {"AttributeName": "title", "AttributeType": "S"},
        ],
        BillingMode="PAY_PER_REQUEST",
    )


class DynamoDBBackend(Backend):
    """
    A graph datastore that uses DynamoDB for persistance and queries.

    """

    def __init__(
        self,
        directed: bool = False,
        node_table_name: str = None,
        edge_table_name: str = None,
        dynamodb_url: str = _DEFAULT_DYNAMODB_URL,
        aws_access_key_id: str = "",
        aws_secret_access_key: str = "",
        primary_key: str = "ID",
    ) -> None:
        """
        Create a new dynamodb-backed graph store.

        Arguments:
            node_table_name (str: "grand_Nodes"): The name to use for the node
                table in DynamoDB.
            edge_table_name (str: "grand_Edges"): The name to use for the edge
                table in DynamoDB.
            dynamodb_url (str: _DEFAULT_DYNAMODB_URL): The URL to use for the
                DynamoDB resource. Defaults to AWS us-east-1.
            primary_key (str: "ID"): The default primary key to use for the
                tables. Note that this key cannot exist in your metadata dicts.

        """
        self._directed = directed
        self._node_table_name = node_table_name or "grand_Nodes"
        self._edge_table_name = edge_table_name or "grand_Edges"

        self._primary_key = primary_key
        self._edge_source_key = "Source"
        self._edge_target_key = "Target"

        self._resource = boto3.resource(
            "dynamodb",
            endpoint_url=dynamodb_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )
        self._client = boto3.client(
            "dynamodb",
            endpoint_url=dynamodb_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )

        if not _dynamo_table_exists(self._node_table_name, self._client):
            node_creation_response = _create_dynamo_table(
                self._node_table_name, self._primary_key, self._resource
            )
            # Await table creation:
            if node_creation_response:
                node_creation_response.meta.client.get_waiter("table_exists").wait(
                    TableName=self._node_table_name
                )

        if not _dynamo_table_exists(self._edge_table_name, self._client):
            edge_creation_response = _create_dynamo_table(
                self._edge_table_name, self._primary_key, self._resource
            )
            # Await table creation:
            if edge_creation_response:
                edge_creation_response.meta.client.get_waiter("table_exists").wait(
                    TableName=self._edge_table_name
                )

        self._node_table = self._resource.Table(self._node_table_name)
        self._edge_table = self._resource.Table(self._edge_table_name)

    def is_directed(self) -> bool:
        """
        Return True if the backend graph is directed.

        Arguments:
            None

        Returns:
            bool: True if the backend graph is directed.

        """
        return self._directed

    def teardown(self, yes_i_am_sure: bool = False):
        """
        Tear down this graph, deleting all evidence it once was here.

        """
        if yes_i_am_sure:
            self._node_table.delete()
            self._edge_table.delete()

    def add_node(self, node_name: Hashable, metadata: dict) -> Hashable:
        """
        Add a new node to the graph.

        Insert a new document into the nodes table.

        Arguments:
            node_name (Hashable): The ID of the node
            metadata (dict: None): An optional dictionary of metadata

        Returns:
            Hashable: The ID of this node, as inserted

        """
        metadata[self._primary_key] = str(node_name)
        response = self._node_table.put_item(Item=metadata)

        return response

    def _scan_table(self, table, scan_kwargs: dict = None):
        done = False
        start_key = None
        results = []
        scan_kwargs = scan_kwargs or {}
        while not done:
            if start_key:
                scan_kwargs["ExclusiveStartKey"] = start_key
            response = table.scan(**scan_kwargs)
            results += response.get("Items", [])
            start_key = response.get("LastEvaluatedKey", None)
            done = start_key is None
        return results

    def all_nodes_as_iterable(self, include_metadata: bool = False) -> Generator:
        """
        Get a generator of all of the nodes in this graph.

        Arguments:
            include_metadata (bool: False): Whether to include node metadata in
                the response

        Returns:
            Generator: A generator of all nodes (arbitrary sort)

        """
        return [
            (
                node[self._primary_key],
                {k: v for k, v in node.items() if k not in [self._primary_key]},
            )
            if include_metadata
            else node[self._primary_key]
            for node in self._scan_table(self._node_table)
        ]

    def has_node(self, u: Hashable) -> bool:
        """
        Return true if the node exists in the graph.

        Arguments:
            u (Hashable): The ID of the node to check

        Returns:
            bool: True if the node exists
        """
        try:
            self._node_table.get_item(u)
            return True
        except:
            return False

    def add_edge(self, u: Hashable, v: Hashable, metadata: dict):
        """
        Add a new edge to the graph between two nodes.

        If the graph is directed, this edge will start (source) at the `u` node
        and end (target) at the `v` node.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID
            metadata (dict): Optional metadata to associate with the edge

        Returns:
            Hashable: The edge ID, as inserted.

        """
        metadata[self._primary_key] = f"__{u}__{v}"
        if self._edge_source_key in metadata:
            raise KeyError(
                f"'{self._edge_source_key}' should not be in metadata. I need that for PK!"
            )
        metadata[self._edge_source_key] = u
        if self._edge_target_key in metadata:
            raise KeyError(
                f"'{self._edge_target_key}' should not be in metadata. I need that for PK!"
            )
        metadata[self._edge_target_key] = v

        if not self.has_node(u):
            self._node_table.put_item(Item={self._primary_key: u})
        if not self.has_node(v):
            self._node_table.put_item(Item={self._primary_key: v})

        response = self._edge_table.put_item(Item=metadata)

        return response

    def all_edges_as_iterable(self, include_metadata: bool = False) -> Generator:
        """
        Get a list of all edges in this graph, arbitrary sort.

        Arguments:
            include_metadata (bool: False): Whether to include edge metadata

        Returns:
            Generator: A generator of all edges (arbitrary sort)

        """
        return [
            (edge[self._edge_source_key], edge[self._edge_target_key], edge)
            if include_metadata
            else (edge[self._edge_source_key], edge[self._edge_target_key])
            for edge in self._scan_table(self._edge_table)
        ]

    def get_node_by_id(self, node_name: Hashable):
        """
        Return the data associated with a node.

        Arguments:
            node_name (Hashable): The node ID to look up

        Returns:
            dict: The metadata associated with this node

        """
        response = self._node_table.get_item(Key={self._primary_key: node_name})

        item = response["Item"]
        item.pop(self._primary_key)
        return item

    def get_edge_by_id(self, u: Hashable, v: Hashable):
        """
        Get an edge by its source and target IDs.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID

        Returns:
            dict: Metadata associated with this edge

        """
        response = self._edge_table.get_item(Key={self._primary_key: f"__{u}__{v}"})
        item = response["Item"]
        item.pop(self._primary_key)
        item.pop(self._edge_source_key)
        item.pop(self._edge_target_key)
        return item

    def get_node_neighbors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        """
        Get a generator of all downstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        if self._directed:
            # Return only edges for which `u` is the source
            res = self._scan_table(
                self._edge_table,
                {"FilterExpression": Key(self._primary_key).begins_with(f"__{u}__"),},
            )

        else:
            res = self._scan_table(
                self._edge_table,
                {
                    "FilterExpression": (
                        Key(self._edge_source_key).eq(u)
                        | Key(self._edge_target_key).eq(u)
                    ),
                },
            )

        if include_metadata:
            results = {}
            for item in res:
                key = (
                    item[self._edge_source_key]
                    if item[self._edge_source_key] != u
                    else item[self._edge_target_key]
                )
                item.pop(self._primary_key)
                item.pop(self._edge_source_key)
                item.pop(self._edge_target_key)
                results[key] = item
            return results
        return iter(
            [
                (
                    edge[self._edge_source_key]
                    if edge[self._edge_source_key] != u
                    else edge[self._edge_target_key]
                )
                for edge in res
            ]
        )

    def get_node_predecessors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        """
        Get a generator of all upstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        if self._directed:
            # Return only edges for which `u` is the target
            res = self._scan_table(
                self._edge_table,
                {"FilterExpression": Key(self._edge_target_key).eq(u),},
            )

        else:
            res = self._scan_table(
                self._edge_table,
                {
                    "FilterExpression": (
                        Key(self._edge_source_key).eq(u)
                        | Key(self._edge_target_key).eq(u)
                    ),
                },
            )

        if include_metadata:
            results = {}
            for item in res:
                key = (
                    item[self._edge_source_key]
                    if item[self._edge_source_key] != u
                    else item[self._edge_target_key]
                )
                item.pop(self._primary_key)
                item.pop(self._edge_source_key)
                item.pop(self._edge_target_key)
                results[key] = item
            return results
        return iter(
            [
                (
                    edge[self._edge_source_key]
                    if edge[self._edge_source_key] != u
                    else edge[self._edge_target_key]
                )
                for edge in res
            ]
        )

    def get_node_count(self) -> Iterable:
        """
        Get an integer count of the number of nodes in this graph.

        Arguments:
            None

        Returns:
            int: The count of nodes

        """
        return self._client.describe_table(TableName=self._node_table_name)["Table"][
            "ItemCount"
        ]

    # Ingesting

    def ingest_from_edgelist_dataframe(
        self, edgelist: pd.DataFrame, source_column: str, target_column: str
    ) -> None:
        """
        Ingest an edgelist from a Pandas DataFrame.

        """
        # Ingest edges first:

        edge_column_names = [
            c for c in edgelist.columns if c not in [source_column, target_column]
        ]

        tic = time.time()
        with self._edge_table.batch_writer() as batch_writer:

            with concurrent.futures.ThreadPoolExecutor(
                _N_PARALLEL_REQUESTS
            ) as executor:
                result_futures = list(
                    map(
                        lambda i: executor.submit(
                            lambda: batch_writer.put_item(
                                Item={
                                    self._primary_key: f"__{edgelist._get_value(i, source_column)}__{edgelist._get_value(i, target_column)}",
                                    self._edge_source_key: edgelist._get_value(
                                        i, source_column
                                    ),
                                    self._edge_target_key: edgelist._get_value(
                                        i, target_column
                                    ),
                                    **{
                                        col: edgelist._get_value(i, col)
                                        for col in edge_column_names
                                    },
                                }
                            )
                        ),
                        edgelist.index,
                    )
                )
                for future in concurrent.futures.as_completed(result_futures):
                    future.result()

            # for edge in edges:
            #     batch_writer.put_item(Item=edge)
        edge_toc = time.time() - tic

        tic = time.time()
        # Construct a unique set of nodes:
        nodes = edgelist[source_column].append(edgelist[target_column]).unique()
        with self._node_table.batch_writer() as batch_writer:
            with concurrent.futures.ThreadPoolExecutor(
                _N_PARALLEL_REQUESTS
            ) as executor:
                result_futures = list(
                    map(
                        lambda x: executor.submit(
                            lambda: batch_writer.put_item(
                                Item={self._primary_key: str(x)}
                            )
                        ),
                        nodes,
                    )
                )
                for future in concurrent.futures.as_completed(result_futures):
                    future.result()

        return {
            "node_count": len(nodes),
            "node_duration": time.time() - tic,
            "edge_count": len(edgelist),
            "edge_duration": edge_toc,
        }



from typing import Hashable, Generator, List, Tuple, Union
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .. import Graph

import pandas as pd

import networkx as nx
from networkx.classes.reportviews import NodeView
from networkx.classes.coreviews import AdjacencyView, AtlasView


class _GrandAdjacencyView(AdjacencyView):

    __slots__ = ("_parent", "_pred_or_succ")  # Still uses AtlasView slots names _atlas

    def __init__(self, parent_nx_dialect: "NetworkXDialect", pred_or_succ: str):
        self._parent = parent_nx_dialect.parent
        self._pred_or_succ = pred_or_succ

    def __getitem__(self, name):
        if self._pred_or_succ == "pred":
            return {
                neighbor: metadata
                for neighbor, metadata in self._parent.backend.get_node_predecessors(
                    name, include_metadata=True
                ).items()
            }
        elif self._pred_or_succ == "succ":
            return {
                neighbor: metadata
                for neighbor, metadata in self._parent.backend.get_node_successors(
                    name, include_metadata=True
                ).items()
            }

    def __len__(self):
        return len(self._parent.backend.get_node_count())

    def __iter__(self):
        return iter(self._parent.backend.all_nodes_as_iterable(include_metadata=False))

    def copy(self):
        raise NotImplementedError()

    def __str__(self):
        return "_GrandAdjacencyView"

    def __repr__(self):
        return "_GrandAdjacencyView"


class _GrandNodeAtlasView(AtlasView):
    def __init__(self, parent):
        self.parent = parent.parent

    def __getitem__(self, key):
        return self.parent.backend.get_node_by_id(key)

    def __len__(self):
        return self.parent.backend.get_node_count()

    def __iter__(self):
        return iter(self.parent.backend.all_nodes_as_iterable(include_metadata=False))

    def copy(self):
        return {
            n: metadata
            for n, metadata in self.parent.backend.all_nodes_as_iterable(
                include_metadata=True
            )
        }

    def __str__(self):
        return "_GrandNodeAtlasView"

    def __repr__(self):
        return "_GrandNodeAtlasView"


class NetworkXDialect(nx.Graph):
    """
    A NetworkXDialect provides a networkx-like interface for graph manipulation

    """

    def __init__(self, parent: "Graph"):
        """
        Create a new dialect to query a backend with NetworkX syntax.

        Arguments:
            parent (Graph): The parent Graph object

        Returns:
            None

        """
        self.parent = parent

    def add_node(self, name: Hashable, **kwargs):
        return self.parent.backend.add_node(name, kwargs)

    def add_edge(self, u: Hashable, v: Hashable, **kwargs):
        return self.parent.backend.add_edge(u, v, kwargs)

    def remove_node(self, name: Hashable):
        if hasattr(self.parent.backend, "remove_node"):
            return self.parent.backend.remove_node(name)
        raise NotImplementedError

    def remove_edge(self, u: Hashable, v: Hashable):
        raise NotImplementedError

    def edges(self, data: bool = False):
        return [
            i for i in self.parent.backend.all_edges_as_iterable(include_metadata=data)
        ]

    def neighbors(self, u: Hashable) -> Generator:
        return self.parent.backend.get_node_neighbors(u)

    def predecessors(self, u: Hashable) -> Generator:
        return self.parent.backend.get_node_predecessors(u)

    def successors(self, u: Hashable) -> Generator:
        return self.parent.backend.get_node_neighbors(u)

    @property
    def _node(self):
        return _GrandNodeAtlasView(self)

    @property
    def adj(self):
        """
        https://github.com/networkx/networkx/blob/master/networkx/classes/digraph.py#L323
        """
        return _GrandAdjacencyView(self, "succ")

    @property
    def _adj(self):
        """
        https://github.com/networkx/networkx/blob/master/networkx/classes/digraph.py#L323
        """
        return _GrandAdjacencyView(self, "succ")

    @property
    def succ(self):
        return _GrandAdjacencyView(self, "succ")

    @property
    def _succ(self):
        return _GrandAdjacencyView(self, "succ")

    @property
    def pred(self):
        """
        https://github.com/networkx/networkx/blob/master/networkx/classes/digraph.py#L323
        """
        return _GrandAdjacencyView(self, "pred")

    @property
    def _pred(self):
        """
        https://github.com/networkx/networkx/blob/master/networkx/classes/digraph.py#L323
        """
        return _GrandAdjacencyView(self, "pred")


class IGraphDialect(nx.Graph):
    """
    An IGraphDialect provides a python-igraph-like interface

    """

    def __init__(self, parent: "Graph"):
        """
        Create a new dialect to query a backend with Python-IGraph syntax.

        Arguments:
            parent (Graph): The parent Graph object

        Returns:
            None

        """
        self.parent = parent

    def add_vertices(self, num_verts: int):
        old_max = len(self.vs)
        for new_v_index in range(num_verts):
            return self.parent.backend.add_node(new_v_index + old_max, {})

    @property
    def vs(self):
        return [
            i for i in self.parent.backend.all_nodes_as_iterable(include_metadata=True)
        ]

    @property
    def es(self):
        return [
            i for i in self.parent.backend.all_edges_as_iterable(include_metadata=True)
        ]

    def add_edges(self, edgelist: List[Tuple[Hashable, Hashable]]):
        for (u, v) in edgelist:
            return self.parent.backend.add_edge(u, v, {})

    def get_edgelist(self):
        return self.parent.backend.all_edges_as_iterable(include_metadata=False)


class NetworkitDialect:
    """
    A Networkit-like API for interacting with a Grand graph.

    For more details on the original API, see here:
    https://networkit.github.io/dev-docs/python_api/graph.html

    """

    def __init__(self, parent: "Graph") -> None:
        self.parent = parent

    def addNode(self):
        new_id = self.parent.backend.get_node_count()
        self.parent.backend.add_node(new_id)
        return new_id

    def addEdge(self, u: Hashable, v: Hashable) -> None:
        self.parent.backend.add_edge(u, v, {})

    def nodes(self):
        return [i for i in self.iterNodes()]

    def iterNodes(self):
        return self.parent.backend.all_nodes_as_iterable()

    def edges(self):
        return [i for i in self.iterEdges()]

    def iterEdges(self):
        return self.parent.backend.all_edges_as_iterable()

    def hasEdge(self, u, v) -> bool:
        return self.parent.backend.get_edge_by_id(u, v) is not None

    def addNodes(self, numberOfNewNodes: int) -> int:
        for _ in range(numberOfNewNodes):
            r = self.addNode()
        return r

    def hasNode(self, u) -> bool:
        return self.parent.backend.has_node(u)

    def degree(self, v):
        return len(self.parent.backend.get_node_neighbors(v))

    def degreeIn(self, v):
        return len(self.parent.backend.get_node_predecessors(v))

    def degreeOut(self, v):
        return len(self.parent.backend.get_node_successors(v))

    def density(self):
        # TODO: implement backend#degree?
        E = len(self.parent.backend.all_edges_as_iterable())
        V = self.parent.backend.get_node_count()

        if self.parent.backend.is_directed():
            return E / (V * (V - 1))
        else:
            return 2 * E / (V * (V - 1))

    def numberOfNodes(self) -> int:
        return self.parent.backend.get_node_count()

    def numberOfEdges(self) -> int:
        return len(self.parent.backend.all_edges_as_iterable())

    def removeEdge(self, u, v) -> None:
        raise NotImplementedError
        return self.parent.backend.remove_edge(u, v)

    def removeNode(self, u: Hashable) -> None:
        if hasattr(self.parent.backend, "remove_node"):
            return self.parent.backend.remove_node(u)
        raise NotImplementedError


    def append(self, G):
        raise NotImplementedError

    def copyNodes(self):
        raise NotImplementedError

    def BFSEdgesFrom(self, start: Union[int, List[int]]):
        raise NotImplementedError


# from dotmotif import dotmotif, NetworkXExecutor
#
# class DotMotifDialect:
#     def __init__(self, parent: "Graph") -> None:
#         """
#         Create a new DotMotifDialect to query a backend with DotMotif syntax.

#         Arguments:
#             parent (Graph): The parent Graph object

#         Returns:
#             None

#         """
#         self.parent = parent

#     def find(
#         self, motif: Union[str, dotmotif], exclude_automorphisms: bool = False
#     ) -> pd.DataFrame:
#         """
#         Find a motif using DotMotif syntax.

#         Arguments:
#             motif (Union[str, dotmotif.dotmotif]): A motif in dotmotif form or
#                 a string in the DotMotif DSL.
#             exclude_automorphisms (bool: True): Whether to exclude motif
#                 automorphisms from the results list

#         Returns:
#             pd.DataFrame: A DataFrame containing the results of the query

#         """
#         if isinstance(motif, str):
#             motif = dotmotif(
#                 ignore_direction=(not self.parent.backend._directed),
#                 exclude_automorphisms=exclude_automorphisms,
#             ).from_motif(motif)

#         return NetworkXExecutor(graph=self.parent.nx).find(motif)

#     def count(
#         self, motif: Union[str, dotmotif], exclude_automorphisms: bool = False
#     ) -> pd.DataFrame:
#         """
#         Count occurrences of a motif using DotMotif syntax.

#         Arguments:
#             motif (Union[str, dotmotif.dotmotif]): A motif in dotmotif form or
#                 a string in the DotMotif DSL.
#             exclude_automorphisms (bool: True): Whether to exclude motif
#                 automorphisms from the results list

#         Returns:
#             int: A count of results of the motif query

#         """
#         if isinstance(motif, str):
#             motif = dotmotif(
#                 ignore_direction=(not self.parent.backend._directed),
#                 exclude_automorphisms=exclude_automorphisms,
#             ).from_motif(motif)

#         return NetworkXExecutor(graph=self.parent.nx).count(motif)



from typing import Hashable, Generator, Iterable
import abc

import networkit
import pandas as pd

from .backend import Backend
from .metadatastore import MetadataStore, DictMetadataStore, NodeNameManager


class NetworkitBackend(Backend):
    """
    Networkit doesn't support metadata or named nodes, so all node names and
    metadata must currently be stored in a parallel data structure.

    To solve this problem, a NodeNameManager and MetadataStore, from
    `grand.backends.metadatastore.NodeNameManager` and
    `grand.backends.metadatastore.MetadataStore` respectively, are included at
    the top level of this class. In order to preserve this metadata structure
    statefully, you must serialize both the graph as well as the data stores.

    This is currently UNOPTIMIZED CODE.

    Recommendations for future work include improved indexing and caching of
    node names and metadata.

    Networkit.graph.Graph documentation:
    https://networkit.github.io/dev-docs/python_api/graph.html

    """

    def __init__(self, directed: bool = False, metadata_store: MetadataStore = None):
        """
        Create a new NetworkitBackend instance, using a Networkit.graph.Graph
        object to store and manage network structure.

        Arguments:
            directed (bool: False): Whether to make the backend graph directed
            metadata_store (MetadataStore): Optionally, a MetadataStore to use
                to handle node and edge attributes. If not provided, defaults
                to a DictMetadataStore.

        Returns:
            None

        """
        self._directed = directed
        self._meta = metadata_store or DictMetadataStore()
        self._nk_graph = networkit.graph.Graph(directed=directed)
        self._names = NodeNameManager()

    def ingest_from_edgelist_dataframe(
        self, edgelist: pd.DataFrame, source_column: str, target_column: str
    ) -> None:
        """
        Ingest an edgelist from a Pandas DataFrame.

        """
        raise NotImplementedError()

    def is_directed(self) -> bool:
        """
        Return True if the backend graph is directed.

        Arguments:
            None

        Returns:
            bool: True if the backend graph is directed.

        """
        return self._directed

    def add_node(self, node_name: Hashable, metadata: dict):
        """
        Add a new node to the graph.

        Arguments:
            node_name (Hashable): The ID of the node
            metadata (dict: None): An optional dictionary of metadata

        Returns:
            Hashable: The ID of this node, as inserted

        """
        # TODO: Remove metadata from lookup if insertion fails
        nk_id = self._nk_graph.addNode()
        self._names.add_node(node_name, nk_id)
        self._meta.add_node(node_name, metadata)
        return nk_id

    def get_node_by_id(self, node_name: Hashable):
        """
        Return the data associated with a node.

        Arguments:
            node_name (Hashable): The node ID to look up

        Returns:
            dict: The metadata associated with this node

        """
        return self._meta.get_node(node_name)

    def all_nodes_as_iterable(self, include_metadata: bool = False) -> Generator:
        """
        Get a generator of all of the nodes in this graph.

        Arguments:
            include_metadata (bool: False): Whether to include node metadata in
                the response

        Returns:
            Generator: A generator of all nodes (arbitrary sort)

        """
        if include_metadata:
            return [
                (self._names.get_name(i), self._meta.get_node(self._names.get_name(i)))
                for i in self._nk_graph.iterNodes()
            ]
        return [self._names.get_name(i) for i in self._nk_graph.iterNodes()]

    def has_node(self, u: Hashable) -> bool:
        """
        Return true if the node exists in the graph.

        Arguments:
            u (Hashable): The ID of the node to check

        Returns:
            bool: True if the node exists
        """
        return u in self._names

    def add_edge(self, u: Hashable, v: Hashable, metadata: dict):
        """
        Add a new edge to the graph between two nodes.

        If the graph is directed, this edge will start (source) at the `u` node
        and end (target) at the `v` node.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID
            metadata (dict): Optional metadata to associate with the edge

        Returns:
            Hashable: The edge ID, as inserted.

        """
        # If u doesn't exist:
        if self.has_node(u):
            x = self._names.get_id(u)
        else:
            x = self.add_node(u, None)

        if self.has_node(v):
            y = self._names.get_id(v)
        else:
            y = self.add_node(v, None)

        # Insert metadata for this edge, replacing the previous metadata:
        self._meta.add_edge(u, v, metadata)

        # TODO: Support multigraphs, and allow duplicate edges.
        if self.has_edge(u, v):
            return
        return self._nk_graph.addEdge(x, y)

    def all_edges_as_iterable(self, include_metadata: bool = False) -> Generator:
        """
        Get a list of all edges in this graph, arbitrary sort.

        Arguments:
            include_metadata (bool: False): Whether to include edge metadata

        Returns:
            Generator: A generator of all edges (arbitrary sort)

        """
        if include_metadata:
            return [
                (
                    self._names.get_name(u),
                    self._names.get_name(v),
                    self._meta.get_edge(
                        self._names.get_name(u), self._names.get_name(v)
                    ),
                )
                for u, v in self._nk_graph.iterEdges()
            ]
        return [
            (self._names.get_name(u), self._names.get_name(v))
            for u, v in self._nk_graph.iterEdges()
        ]

    def has_edge(self, u, v):
        return self._nk_graph.hasEdge(self._names.get_id(u), self._names.get_id(v))

    def get_edge_by_id(self, u: Hashable, v: Hashable):
        """
        Get an edge by its source and target IDs.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID

        Returns:
            dict: Metadata associated with this edge

        """
        if self.has_edge(u, v):
            return self._meta.get_edge(u, v)
        raise IndexError(f"The edge ({u}, {v}) is not in the graph.")

    def get_node_successors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        return self.get_node_neighbors(u, include_metadata)

    def get_node_neighbors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        """
        Get a generator of all downstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        my_id = self._names.get_id(u)
        if include_metadata:
            val = {}
            for vid in self._nk_graph.iterNeighbors(my_id):
                v = self._names.get_name(vid)
                if self.is_directed():
                    val[v] = self._meta.get_edge(u, v)
                else:
                    try:
                        val[v] = self._meta.get_edge(u, v)
                    except KeyError:
                        val[v] = self._meta.get_edge(v, u)
            return val

        return iter(
            [self._names.get_name(i) for i in self._nk_graph.iterNeighbors(my_id)]
        )

    def get_node_predecessors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        """
        Get a generator of all upstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        my_id = self._names.get_id(u)
        if include_metadata:
            val = {}
            for vid in self._nk_graph.iterInNeighbors(my_id):
                v = self._names.get_name(vid)
                if self.is_directed():
                    val[v] = self._meta.get_edge(v, u)
                else:
                    try:
                        val[v] = self._meta.get_edge(u, v)
                    except KeyError:
                        val[v] = self._meta.get_edge(v, u)
            return val

        return iter(
            [self._names.get_name(i) for i in self._nk_graph.iterInNeighbors(my_id)]
        )

    def get_node_count(self) -> Iterable:
        """
        Get an integer count of the number of nodes in this graph.

        Arguments:
            None

        Returns:
            int: The count of nodes

        """
        return len([i for i in self.all_nodes_as_iterable()])



from typing import Hashable, Iterable
import abc

import pandas as pd


class Backend(abc.ABC):
    """
    Abstract base class for the management of persisted graph structure.

    Do not use this class directly.

    """

    def __init__(self, directed: bool = False):
        """
        Create a new Backend instance.

        Arguments:
            directed (bool: False): Whether to make the backend graph directed

        Returns:
            None

        """
        ...

    def ingest_from_edgelist_dataframe(
        self, edgelist: pd.DataFrame, source_column: str, target_column: str
    ) -> None:
        """
        Ingest an edgelist from a Pandas DataFrame.

        """
        ...

    def is_directed(self) -> bool:
        """
        Return True if the backend graph is directed.

        Arguments:
            None

        Returns:
            bool: True if the backend graph is directed.

        """
        ...

    def add_node(self, node_name: Hashable, metadata: dict):
        """
        Add a new node to the graph.

        Arguments:
            node_name (Hashable): The ID of the node
            metadata (dict: None): An optional dictionary of metadata

        Returns:
            Hashable: The ID of this node, as inserted

        """
        ...

    def get_node_by_id(self, node_name: Hashable):
        """
        Return the data associated with a node.

        Arguments:
            node_name (Hashable): The node ID to look up

        Returns:
            dict: The metadata associated with this node

        """
        ...

    def all_nodes_as_iterable(self, include_metadata: bool = False) -> Iterable:
        """
        Get a generator of all of the nodes in this graph.

        Arguments:
            include_metadata (bool: False): Whether to include node metadata in
                the response

        Returns:
            Generator: A generator of all nodes (arbitrary sort)

        """
        ...

    def has_node(self, u: Hashable) -> bool:
        """
        Return true if the node exists in the graph.

        Arguments:
            u (Hashable): The ID of the node to check

        Returns:
            bool: True if the node exists
        """
        ...

    def add_edge(self, u: Hashable, v: Hashable, metadata: dict):
        """
        Add a new edge to the graph between two nodes.

        If the graph is directed, this edge will start (source) at the `u` node
        and end (target) at the `v` node.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID
            metadata (dict): Optional metadata to associate with the edge

        Returns:
            Hashable: The edge ID, as inserted.

        """
        ...

    def all_edges_as_iterable(self, include_metadata: bool = False) -> Iterable:
        """
        Get a list of all edges in this graph, arbitrary sort.

        Arguments:
            include_metadata (bool: False): Whether to include edge metadata

        Returns:
            Generator: A generator of all edges (arbitrary sort)

        """
        ...

    def get_edge_by_id(self, u: Hashable, v: Hashable):
        """
        Get an edge by its source and target IDs.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID

        Returns:
            dict: Metadata associated with this edge

        """
        ...

    def get_node_successors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Iterable:
        return self.get_node_neighbors(u, include_metadata)

    def get_node_neighbors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Iterable:
        """
        Get a generator of all downstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        ...

    def get_node_predecessors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Iterable:
        """
        Get a generator of all upstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        ...

    def get_node_count(self) -> Iterable:
        """
        Get an integer count of the number of nodes in this graph.

        Arguments:
            None

        Returns:
            int: The count of nodes

        """
        return len([i for i in self.all_nodes_as_iterable()])



"""
https://tinkerpop.apache.org/docs/current/reference/
"""

from typing import Hashable, Generator, Iterable
import time

import pandas as pd
from gremlin_python.structure.graph import Graph
from gremlin_python.process.graph_traversal import __, GraphTraversalSource
from gremlin_python.driver.driver_remote_connection import DriverRemoteConnection

from .backend import Backend

ID = "__id"
EDGE_NAME = "__edge"
NODE_NAME = "__node"


def _node_to_metadata(n):
    return {k if isinstance(k, str) else k.name: v for k, v in n.items()}


class GremlinBackend(Backend):
    """
    A backend instance for Gremlin-compatible graph databases.

    """

    def __init__(self, graph: GraphTraversalSource, directed: bool = True):
        """
        Create a new Backend instance wrapping a Gremlin endpoint.

        Arguments:
            directed (bool: False): Whether to make the backend graph directed

        Returns:
            None

        """
        self._g = graph

    def is_directed(self) -> bool:
        """
        Return True if the backend graph is directed.

        The Gremlin-backed datastore is always directed.

        Arguments:
            None

        Returns:
            bool: True if the backend graph is directed.

        """
        return True

    def add_node(self, node_name: Hashable, metadata: dict):
        """
        Add a new node to the graph.

        Arguments:
            node_name (Hashable): The ID of the node
            metadata (dict: None): An optional dictionary of metadata

        Returns:
            Hashable: The ID of this node, as inserted

        """
        if self.has_node(node_name):
            # Retrieve the existing node; we will update the props.
            v = self._g.V().has(ID, node_name)
        else:
            v = self._g.addV().property(ID, node_name)
        for key, val in metadata.items():
            v = v.property(key, val)
        return v.toList()[0]

    def get_node_by_id(self, node_name: Hashable):
        """
        Return the data associated with a node.

        Arguments:
            node_name (Hashable): The node ID to look up

        Returns:
            dict: The metadata associated with this node

        """
        try:
            return _node_to_metadata(
                self._g.V().has(ID, node_name).valueMap(True).toList()[0]
            )
        except IndexError as e:
            raise KeyError() from e

    def has_node(self, u: Hashable) -> bool:
        """
        Return the data associated with a node.

        Arguments:
            node_name (Hashable): The node ID to look up

        Returns:
            dict: The metadata associated with this node

        """
        try:
            self.get_node_by_id(u)
            return True
        except KeyError:
            return False

    def remove_node(self, node_name: Hashable):
        """
        Remove a node.

        Arguments:
            node_name (Hashable): The node ID to look up

        Returns:
            dict: The metadata associated with this node

        """
        return self._g.V().has(ID, node_name).drop().toList()

    def all_nodes_as_iterable(self, include_metadata: bool = False) -> Generator:
        """
        Get a generator of all of the nodes in this graph.

        Arguments:
            include_metadata (bool: False): Whether to include node metadata in
                the response

        Returns:
            Generator: A generator of all nodes (arbitrary sort)

        """
        if include_metadata:
            return iter(
                [
                    {n[ID][0]: _node_to_metadata(n)}
                    for n in self._g.V().valueMap(True).toList()
                ]
            )
        else:
            return iter([n[ID] for n in self._g.V().project(ID).by(ID).toList()])

    def add_edge(self, u: Hashable, v: Hashable, metadata: dict):
        """
        Add a new edge to the graph between two nodes.

        If the graph is directed, this edge will start (source) at the `u` node
        and end (target) at the `v` node.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID
            metadata (dict): Optional metadata to associate with the edge

        Returns:
            Hashable: The edge ID, as inserted.

        """
        try:
            self.get_edge_by_id(u, v)
            e = self._g.V().has(ID, u).outE().as_("e").inV().has(ID, v).select("e")
        except IndexError:
            if not self.has_node(u):
                self.add_node(u, {})
            if not self.has_node(v):
                self.add_node(v, {})
            e = (
                self._g.V()
                .has(ID, u)
                .addE(EDGE_NAME)
                .as_("e")
                .to(__.V().has(ID, v))
                .select("e")
            )
        for key, val in metadata.items():
            e = e.property(key, val)
        return e.toList()

    def all_edges_as_iterable(self, include_metadata: bool = False) -> Generator:
        """
        Get a list of all edges in this graph, arbitrary sort.

        Arguments:
            include_metadata (bool: False): Whether to include edge metadata

        Returns:
            Generator: A generator of all edges (arbitrary sort)

        """
        if include_metadata:
            return iter(
                [
                    (e["source"], e["target"], _node_to_metadata(e["properties"]))
                    for e in (
                        self._g.V()
                        .outE()
                        .project("target", "source", "properties")
                        .by(__.inV().values(ID))
                        .by(__.outV().values(ID))
                        .by(__.valueMap(True))
                        .toList()
                    )
                ]
            )
        return iter(
            [
                (e["source"], e["target"])
                for e in self._g.V()
                .outE()
                .project("target", "source")
                .by(__.inV().values(ID))
                .by(__.outV().values(ID))
                .toList()
            ]
        )

    def get_edge_by_id(self, u: Hashable, v: Hashable):
        """
        Get an edge by its source and target IDs.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID

        Returns:
            dict: Metadata associated with this edge

        """
        return (
            self._g.V()
            .has(ID, u)
            .outE()
            .as_("e")
            .inV()
            .has(ID, v)
            .select("e")
            .properties()
            .toList()
        )[0]

    def get_node_neighbors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        """
        Get a generator of all downstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        if include_metadata:
            return {
                e["target"]: _node_to_metadata(e["properties"])
                for e in (
                    self._g.V()
                    .has(ID, u)
                    .outE()
                    .project("target", "source", "properties")
                    .by(__.inV().values(ID))
                    .by(__.outV().values(ID))
                    .by(__.valueMap(True))
                    .toList()
                )
            }
        return self._g.V().has(ID, u).out().values(ID).toList()

    def get_node_predecessors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        """
        Get a generator of all downstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        if include_metadata:
            return {
                e["source"]: e
                for e in (
                    self._g.V()
                    .has(ID, u)
                    .inE()
                    .project("target", "source", "properties")
                    .by(__.inV().values(ID))
                    .by(__.outV().values(ID))
                    .by(__.valueMap(True))
                    .toList()
                )
            }
        return self._g.V().out().has(ID, u).values(ID).toList()

    def get_node_count(self) -> Iterable:
        """
        Get an integer count of the number of nodes in this graph.

        Arguments:
            None

        Returns:
            int: The count of nodes

        """
        return self._g.V().count().toList()[0]

    def teardown(self) -> None:
        self._g.V().drop().toList()



import pytest
import os

import networkx as nx

from . import NetworkXBackend, SQLBackend, DynamoDBBackend
from .. import Graph

backend_test_params = [
    pytest.param(
        NetworkXBackend,
        marks=pytest.mark.skipif(
            os.environ.get("TEST_NETWORKXBACKEND", default="1") != "1",
            reason="NetworkX Backend skipped because $TEST_NETWORKXBACKEND != 0.",
        ),
    ),
    pytest.param(
        SQLBackend,
        marks=pytest.mark.skipif(
            os.environ.get("TEST_SQLBACKEND", default="1") != "1",
            reason="SQL Backend skipped because $TEST_SQLBACKEND != 0.",
        ),
    ),
    pytest.param(
        DynamoDBBackend,
        marks=pytest.mark.skipif(
            os.environ.get("TEST_DYNAMODBBACKEND") != "1",
            reason="DynamoDB Backend skipped because $TEST_DYNAMODBBACKEND != 0.",
        ),
    ),
]

if os.environ.get("TEST_NETWORKITBACKEND") == "1":
    from .networkit import NetworkitBackend

    backend_test_params.append(
        pytest.param(
            NetworkitBackend,
            marks=pytest.mark.skipif(
                os.environ.get("TEST_NETWORKITBACKEND") != "1",
                reason="Networkit Backend skipped because $TEST_NETWORKITBACKEND != 0.",
            ),
        ),
    )

if os.environ.get("TEST_IGRAPHBACKEND") == "1":
    from .igraph import IGraphBackend

    backend_test_params.append(
        pytest.param(
            IGraphBackend,
            marks=pytest.mark.skipif(
                os.environ.get("TEST_IGRAPHBACKEND") != "1",
                reason="Networkit Backend skipped because $TEST_IGRAPHBACKEND != 0.",
            ),
        ),
    )


@pytest.mark.parametrize("backend", backend_test_params)
class TestBackend:
    def test_can_create(self, backend):
        backend()

    def test_can_add_node(self, backend):
        G = Graph(backend=backend())
        nxG = nx.Graph()
        G.nx.add_node("A", k="v")
        nxG.add_node("A", k="v")
        assert len(G.nx.nodes()) == len(nxG.nodes())
        G.nx.add_node("B", k="v")
        nxG.add_node("B", k="v")
        assert len(G.nx.nodes()) == len(nxG.nodes())

    def test_can_add_edge(self, backend):
        G = Graph(backend=backend())
        nxG = nx.Graph()
        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        assert len(G.nx.edges()) == len(nxG.edges())
        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        assert len(G.nx.edges()) == len(nxG.edges())

    def test_can_get_node(self, backend):
        G = Graph(backend=backend())
        nxG = nx.Graph()
        md = dict(k="B")
        G.nx.add_node("A", **md)
        nxG.add_node("A", **md)
        assert G.nx.nodes["A"] == nxG.nodes["A"]

    def test_can_get_edge(self, backend):
        G = Graph(backend=backend())
        nxG = nx.Graph()
        md = {"k":"B"}
        G.nx.add_edge("A", "B", **md)
        nxG.add_edge("A", "B", **md)
        assert G.nx.get_edge_data("A", "B") == nxG.get_edge_data("A", "B")

    def test_can_get_neighbors(self, backend):
        G = Graph(backend=backend())
        nxG = nx.Graph()
        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        assert sorted([i for i in G.nx.neighbors("A")]) == sorted(
            [i for i in nxG.neighbors("A")]
        )
        assert sorted([i for i in G.nx.neighbors("B")]) == sorted(
            [i for i in nxG.neighbors("B")]
        )
        G.nx.add_edge("A", "C")
        nxG.add_edge("A", "C")
        assert sorted([i for i in G.nx.neighbors("A")]) == sorted(
            [i for i in nxG.neighbors("A")]
        )
        assert sorted([i for i in G.nx.neighbors("B")]) == sorted(
            [i for i in nxG.neighbors("B")]
        )
        assert sorted([i for i in G.nx.neighbors("C")]) == sorted(
            [i for i in nxG.neighbors("C")]
        )

    def test_undirected_adj(self, backend):
        G = Graph(backend=backend())
        nxG = nx.Graph()
        assert G.nx._adj == nxG._adj
        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        assert G.nx._adj == nxG._adj

    def test_directed_adj(self, backend):
        G = Graph(backend=SQLBackend(directed=True))
        nxG = nx.DiGraph()
        assert G.nx._adj == nxG._adj
        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        assert G.nx._adj == nxG._adj

    def test_can_traverse_undirected_graph(self, backend):
        G = Graph(backend=backend())
        nxG = nx.Graph()
        md = dict(k="B")
        G.nx.add_edge("A", "B", **md)
        nxG.add_edge("A", "B", **md)
        assert dict(nx.bfs_successors(G.nx, "A")) == dict(nx.bfs_successors(nxG, "A"))
        G.nx.add_edge("B", "C", **md)
        nxG.add_edge("B", "C", **md)
        assert dict(nx.bfs_successors(G.nx, "A")) == dict(nx.bfs_successors(nxG, "A"))
        G.nx.add_edge("B", "D", **md)
        nxG.add_edge("B", "D", **md)
        assert dict(nx.bfs_successors(G.nx, "A")) == dict(nx.bfs_successors(nxG, "A"))
        assert dict(nx.bfs_successors(G.nx, "C")) == dict(nx.bfs_successors(nxG, "C"))

    def test_can_traverse_directed_graph(self, backend):
        G = Graph(backend=backend(directed=True))
        nxG = nx.DiGraph()
        md = dict(k="B")
        G.nx.add_edge("A", "B", **md)
        nxG.add_edge("A", "B", **md)
        assert dict(nx.bfs_successors(G.nx, "A")) == dict(nx.bfs_successors(nxG, "A"))
        G.nx.add_edge("B", "C", **md)
        nxG.add_edge("B", "C", **md)
        assert dict(nx.bfs_successors(G.nx, "A")) == dict(nx.bfs_successors(nxG, "A"))
        G.nx.add_edge("B", "D", **md)
        nxG.add_edge("B", "D", **md)
        assert dict(nx.bfs_successors(G.nx, "A")) == dict(nx.bfs_successors(nxG, "A"))
        assert dict(nx.bfs_successors(G.nx, "C")) == dict(nx.bfs_successors(nxG, "C"))

    def test_subgraph_isomorphism_undirected(self, backend):
        G = Graph(backend=backend(directed=False))
        nxG = nx.Graph()

        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        G.nx.add_edge("B", "C")
        nxG.add_edge("B", "C")
        G.nx.add_edge("C", "A")
        nxG.add_edge("C", "A")

        from networkx.algorithms.isomorphism import GraphMatcher

        assert len(
            [i for i in GraphMatcher(G.nx, G.nx).subgraph_monomorphisms_iter()]
        ) == len([i for i in GraphMatcher(nxG, nxG).subgraph_monomorphisms_iter()])

    def test_subgraph_isomorphism_directed(self, backend):
        G = Graph(backend=backend(directed=True))
        nxG = nx.DiGraph()

        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        G.nx.add_edge("B", "C")
        nxG.add_edge("B", "C")
        G.nx.add_edge("C", "A")
        nxG.add_edge("C", "A")

        from networkx.algorithms.isomorphism import DiGraphMatcher

        assert len(
            [i for i in DiGraphMatcher(G.nx, G.nx).subgraph_monomorphisms_iter()]
        ) == len([i for i in DiGraphMatcher(nxG, nxG).subgraph_monomorphisms_iter()])

    def test_can_get_edge_metadata(self, backend):
        G = Graph(backend=backend())
        G.nx.add_edge("foo", "bar", baz=True)
        assert G.nx.edges(data=True) == [("foo", "bar", {"baz": True})]

    def test_edge_dne_raises(self, backend):
        G = Graph(backend=backend())
        G.nx.add_edge("foo", "bar", baz=True)

        assert G.nx.has_edge("foo", "crab") == False
        assert G.nx.has_edge("foo", "bar") == True
        # assert G.nx.edges[("foo", "bar")] != None
        # with pytest.raises(Exception):
        #     G.nx.edges[("foo", "crab")]

    def test_reverse_edges_in_undirected(self, backend):
        G = Graph(backend=backend(directed=False), directed=False)
        G.nx.add_edge("foo", "bar", baz=True)

        assert G.nx.has_edge("foo", "bar") == True
        assert G.nx.has_edge("bar", "foo") == True



"""
Grand graph databasifier.

Aug 2020
"""

from .backends import Backend, NetworkXBackend
from .dialects import NetworkXDialect, IGraphDialect


_DEFAULT_BACKEND = NetworkXBackend

__version__ = "0.2.0"


class Graph:
    """
    A grand.Graph enables you to manipulate a graph using multiple dialects.

    """

    def __init__(
        self, backend: Backend = None, directed: bool = True, dialects: dict = None
    ):
        """
        Create a new grand.Graph.

        Arguments:
            backend (Backend): The backend to use. If none is provided, will
                default to _DEFAULT_BACKEND.

        Returns:
            None

        """
        self.backend = backend or _DEFAULT_BACKEND(directed=directed)

        # Attach dialects:
        self.nx = NetworkXDialect(self)
        self.igraph = IGraphDialect(self)

        if dialects:
            self.dialects = {k: v(self) for k, v in dialects.items()}

    def save(self, filename: str) -> str:
        raise NotImplementedError()

    @staticmethod
    def load(self, filename: str) -> str:
        raise NotImplementedError()



import unittest

from .. import Graph
import networkx as nx


class TestIGraphDialect(unittest.TestCase):
    def test_igraph_verts(self):
        G = Graph()
        self.assertEqual(G.igraph.vs, [])
        G.nx.add_node("1")
        self.assertEqual(G.igraph.vs, [("1", {})])


class TestNetworkXDialect(unittest.TestCase):
    def test_nx_pred(self):
        G = Graph(directed=True)
        G.nx.add_edge("1", "2")
        G.nx.add_edge("1", "3")
        H = nx.DiGraph()
        H.add_edge("1", "2")
        H.add_edge("1", "3")
        self.assertEqual(G.nx.pred, H.pred)


# class TestDotMotifDialect(unittest.TestCase):
#     def test_dm_monomorphism_undirected(self):
#         G = Graph(directed=False)
#         G.nx.add_edge("1", "2")
#         G.nx.add_edge("2", "3")
#         G.nx.add_edge("3", "1")

#         self.assertEqual(
#             len(G.dm.find("""A -> B\nB -> C\nC -> A""")), 6,
#         )

#     def test_dm_monomorphism_directed(self):
#         G = Graph(directed=True)
#         G.nx.add_edge("1", "2")
#         G.nx.add_edge("2", "3")
#         G.nx.add_edge("3", "1")

#         self.assertEqual(
#             len(G.dm.find("""A -> B\nB -> C\nC -> A""")), 3,
#         )

#     def test_dm_monomorphism_undirected_automorphisms(self):
#         G = Graph(directed=False)
#         G.nx.add_edge("1", "2")
#         G.nx.add_edge("2", "3")
#         G.nx.add_edge("3", "1")
#         G.nx.add_edge("3", "4")

#         self.assertEqual(
#             len(G.dm.find("""A -> B\nB -> C\nC -> A""", exclude_automorphisms=True)), 1,
#         )

#     def test_count_dm_monomorphism_undirected(self):
#         G = Graph(directed=False)
#         G.nx.add_edge("1", "2")
#         G.nx.add_edge("2", "3")
#         G.nx.add_edge("3", "1")

#         self.assertEqual(
#             G.dm.count("""A -> B\nB -> C\nC -> A"""), 6,
#         )

#     def test_count_dm_monomorphism_directed(self):
#         G = Graph(directed=True)
#         G.nx.add_edge("1", "2")
#         G.nx.add_edge("2", "3")
#         G.nx.add_edge("3", "1")

#         self.assertEqual(
#             G.dm.count("""A -> B\nB -> C\nC -> A"""), 3,
#         )

#     def test_count_dm_monomorphism_undirected_automorphisms(self):
#         G = Graph(directed=False)
#         G.nx.add_edge("1", "2")
#         G.nx.add_edge("2", "3")
#         G.nx.add_edge("3", "1")
#         G.nx.add_edge("3", "4")

#         self.assertEqual(
#             G.dm.count("""A -> B\nB -> C\nC -> A""", exclude_automorphisms=True), 1,
#         )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

["sqlalchemy"]

setuptools.setup(
    name="grand-graph",
    version="0.2.0",
    author="Jordan Matelsky",
    author_email="opensource@matelsky.com",
    description="Graph database wrapper for non-graph datastores",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aplbrain/grand",
    packages=setuptools.find_packages(),
    install_requires=[
        "boto3",
        "networkx==2.4",
        "numpy==1.19.1",
        "pandas==1.1.0",
        "SQLAlchemy==1.3.18",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from typing import Hashable


class MetadataStore:
    def add_node(self, node_name: Hashable, metadata: dict) -> Hashable:
        raise NotImplementedError()

    def add_edge(self, u: Hashable, v: Hashable, metadata: dict) -> Hashable:
        raise NotImplementedError()

    def get_node(self, node_name: Hashable) -> dict:
        raise NotImplementedError()

    def get_edge(self, u: Hashable, v: Hashable) -> dict:
        raise NotImplementedError()


class DictMetadataStore(MetadataStore):
    def __init__(self):
        self.ndata = {}
        self.edata = {}

    def add_node(self, node_name: Hashable, metadata: dict) -> Hashable:
        if metadata is None:
            metadata = {}
        self.ndata[node_name] = metadata

    def add_edge(self, u: Hashable, v: Hashable, metadata: dict) -> Hashable:
        if metadata is None:
            metadata = {}
        self.edata[(u, v)] = metadata

    def get_node(self, node_name: Hashable) -> dict:
        return self.ndata[node_name]

    def get_edge(self, u: Hashable, v: Hashable) -> dict:
        return self.edata[(u, v)]


class NodeNameManager:
    def __init__(self):
        self.node_names_by_id = {}
        self.node_ids_by_name = {}

    def add_node(self, name: Hashable, _id: Hashable):
        self.node_names_by_id[_id] = name
        self.node_ids_by_name[name] = _id

    def get_name(self, _id: Hashable) -> Hashable:
        return self.node_names_by_id[_id]

    def get_id(self, name: Hashable) -> Hashable:
        return self.node_ids_by_name[name]

    def __contains__(self, name: Hashable) -> bool:
        return name in self.node_ids_by_name



from typing import Hashable, Generator, Iterable

from igraph import Graph, InternalError
import pandas as pd

from .backend import Backend


def _remove_name_from_attributes(attributes: dict):
    attrs = attributes
    attrs.pop("name")
    return attrs


class IGraphBackend(Backend):
    """
    This is currently UNOPTIMIZED CODE.

    Recommendations for future work include improved indexing and caching of
    node names and metadata.

    """

    def __init__(self, directed: bool = False):
        """
        Create a new IGraphBackend instance, using an igraph.Graph object to
        store and manage network structure.

        Arguments:
            directed (bool: False): Whether to make the backend graph directed

        Returns:
            None

        """
        self._directed = directed
        self._ig = Graph(directed=self._directed)

    def ingest_from_edgelist_dataframe(
        self, edgelist: pd.DataFrame, source_column: str, target_column: str
    ) -> None:
        """
        Ingest an edgelist from a Pandas DataFrame.

        """
        raise NotImplementedError()

    def is_directed(self) -> bool:
        """
        Return True if the backend graph is directed.

        Arguments:
            None

        Returns:
            bool: True if the backend graph is directed.

        """
        return self._directed

    def add_node(self, node_name: Hashable, metadata: dict):
        """
        Add a new node to the graph.

        Arguments:
            node_name (Hashable): The ID of the node
            metadata (dict: None): An optional dictionary of metadata

        Returns:
            Hashable: The ID of this node, as inserted

        """
        # TODO: overwrite existing
        self._ig.add_vertex(node_name, **metadata)
        return node_name

    def get_node_by_id(self, node_name: Hashable):
        """
        Return the data associated with a node.

        Arguments:
            node_name (Hashable): The node ID to look up

        Returns:
            dict: The metadata associated with this node

        """
        return _remove_name_from_attributes(
            self._ig.vs.find(name=node_name).attributes()
        )

    def all_nodes_as_iterable(self, include_metadata: bool = False) -> Generator:
        """
        Get a generator of all of the nodes in this graph.

        Arguments:
            include_metadata (bool: False): Whether to include node metadata in
                the response

        Returns:
            Generator: A generator of all nodes (arbitrary sort)

        """
        if include_metadata:
            for v in self._ig.vs:
                yield (v["name"], _remove_name_from_attributes(v.attributes()))
        else:
            for i in self._ig.vs:
                yield i["name"]

    def has_node(self, u: Hashable) -> bool:
        """
        Return true if the node exists in the graph.

        Arguments:
            u (Hashable): The ID of the node to check

        Returns:
            bool: True if the node exists
        """
        try:
            self._ig.vs.find(name=u)
            return True
        except:
            return False

    def add_edge(self, u: Hashable, v: Hashable, metadata: dict):
        """
        Add a new edge to the graph between two nodes.

        If the graph is directed, this edge will start (source) at the `u` node
        and end (target) at the `v` node.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID
            metadata (dict): Optional metadata to associate with the edge

        Returns:
            Hashable: The edge ID, as inserted.

        """
        if self.has_edge(u, v):
            # TODO: update metadata
            return
        if not self.has_node(u):
            self.add_node(u, {})
        if not self.has_node(v):
            self.add_node(v, {})
        return self._ig.add_edge(source=u, target=v, **metadata)

    def all_edges_as_iterable(self, include_metadata: bool = False) -> Generator:
        """
        Get a list of all edges in this graph, arbitrary sort.

        Arguments:
            include_metadata (bool: False): Whether to include edge metadata

        Returns:
            Generator: A generator of all edges (arbitrary sort)

        """
        if include_metadata:
            for e in self._ig.es:
                yield (e.source_vertex["name"], e.target_vertex["name"], e.attributes())
        else:
            for e in self._ig.es:
                yield e.source_vertex["name"], e.target_vertex["name"]

    def has_edge(self, u, v):
        try:
            self._ig.get_eid(u, v)
            return True
        except (InternalError, ValueError):
            # InternalError means no such edge
            # ValueError means one vertex doesn't exist
            return False

    def get_edge_by_id(self, u: Hashable, v: Hashable):
        """
        Get an edge by its source and target IDs.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID

        Returns:
            dict: Metadata associated with this edge

        """
        try:
            return self._ig.es[self._ig.get_eid(u, v)].attributes()
        except InternalError:
            raise IndexError(f"The edge ({u}, {v}) is not in the graph.")

    def get_node_successors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        return self.get_node_neighbors(u, include_metadata)

    def get_node_neighbors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        """
        Get a generator of all downstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        if include_metadata:
            return {
                self._ig.vs[s]["name"]: self.get_edge_by_id(u, s)
                for s in self._ig.successors(u)
            }
        else:
            return iter([self._ig.vs[s]["name"] for s in self._ig.successors(u)])

    def get_node_predecessors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        """
        Get a generator of all upstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        if include_metadata:
            return {
                self._ig.vs[s]["name"]: self.get_edge_by_id(s, u)
                for s in self._ig.predecessors(u)
            }
        else:
            return iter([self._ig.vs[s]["name"] for s in self._ig.predecessors(u)])

    def get_node_count(self) -> Iterable:
        """
        Get an integer count of the number of nodes in this graph.

        Arguments:
            None

        Returns:
            int: The count of nodes

        """
        return self._ig.vcount()



from typing import Hashable, Generator, Iterable
import time

import boto3
import pandas as pd
import networkx as nx

from .backend import Backend


class NetworkXBackend(Backend):
    def __init__(self, directed: bool = False):
        """
        Create a new Backend instance.

        Arguments:
            directed (bool: False): Whether to make the backend graph directed

        Returns:
            None

        """
        self._nx_graph = nx.DiGraph() if directed else nx.Graph()
        self._directed = directed

    def is_directed(self) -> bool:
        """
        Return True if the backend graph is directed.

        Arguments:
            None

        Returns:
            bool: True if the backend graph is directed.

        """
        return self._directed

    def add_node(self, node_name: Hashable, metadata: dict):
        """
        Add a new node to the graph.

        Arguments:
            node_name (Hashable): The ID of the node
            metadata (dict: None): An optional dictionary of metadata

        Returns:
            Hashable: The ID of this node, as inserted

        """
        self._nx_graph.add_node(node_name, **metadata)

    def get_node_by_id(self, node_name: Hashable):
        """
        Return the data associated with a node.

        Arguments:
            node_name (Hashable): The node ID to look up

        Returns:
            dict: The metadata associated with this node

        """
        return self._nx_graph.nodes[node_name]

    def all_nodes_as_iterable(self, include_metadata: bool = False) -> Generator:
        """
        Get a generator of all of the nodes in this graph.

        Arguments:
            include_metadata (bool: False): Whether to include node metadata in
                the response

        Returns:
            Generator: A generator of all nodes (arbitrary sort)

        """
        return self._nx_graph.nodes(data=include_metadata)

    def add_edge(self, u: Hashable, v: Hashable, metadata: dict):
        """
        Add a new edge to the graph between two nodes.

        If the graph is directed, this edge will start (source) at the `u` node
        and end (target) at the `v` node.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID
            metadata (dict): Optional metadata to associate with the edge

        Returns:
            Hashable: The edge ID, as inserted.

        """
        self._nx_graph.add_edge(u, v, **metadata)

    def all_edges_as_iterable(self, include_metadata: bool = False) -> Generator:
        """
        Get a list of all edges in this graph, arbitrary sort.

        Arguments:
            include_metadata (bool: False): Whether to include edge metadata

        Returns:
            Generator: A generator of all edges (arbitrary sort)

        """
        return self._nx_graph.edges(data=include_metadata)

    def get_edge_by_id(self, u: Hashable, v: Hashable):
        """
        Get an edge by its source and target IDs.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID

        Returns:
            dict: Metadata associated with this edge

        """
        return self._nx_graph.edges[u, v]

    def get_node_neighbors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        """
        Get a generator of all downstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        if include_metadata:
            return self._nx_graph[u]
        return self._nx_graph.neighbors(u)

    def get_node_predecessors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Generator:
        """
        Get a generator of all downstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        if include_metadata:
            return self._nx_graph.pred[u]
        return self._nx_graph.predecessors(u)

    def get_node_count(self) -> Iterable:
        """
        Get an integer count of the number of nodes in this graph.

        Arguments:
            None

        Returns:
            int: The count of nodes

        """
        return len(self._nx_graph)

    def ingest_from_edgelist_dataframe(
        self, edgelist: pd.DataFrame, source_column: str, target_column: str
    ) -> None:
        """
        Ingest an edgelist from a Pandas DataFrame.

        """

        tic = time.time()
        self._nx_graph.add_edges_from(
            [
                (
                    e[source_column],
                    e[target_column],
                    {
                        k: v
                        for k, v in dict(e).items()
                        if k not in [source_column, target_column]
                    },
                )
                for _, e in edgelist.iterrows()
            ]
        )

        nodes = edgelist[source_column].append(edgelist[target_column]).unique()

        return {
            "node_count": len(nodes),
            "node_duration": 0,
            "edge_count": len(edgelist),
            "edge_duration": time.time() - tic,
        }

    def teardown(self) -> None:
        return
